@@ -21,32 +21,41 @@
  * Or visit:  http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
  */
 
+#include "config.h"
+
 #include "loadsofa.h"
 
 #include <algorithm>
 #include <atomic>
 #include <chrono>
 #include <cmath>
-#include <cstdio>
 #include <functional>
 #include <future>
+#include <iostream>
 #include <iterator>
 #include <memory>
 #include <optional>
+#include <span>
 #include <string>
 #include <string_view>
 #include <thread>
 #include <vector>
 
-#include "alspan.h"
 #include "alnumeric.h"
-#include "fmt/core.h"
+#include "fmt/base.h"
+#include "fmt/ostream.h"
 #include "makemhr.h"
 #include "polyphase_resampler.h"
 #include "sofa-support.h"
 
 #include "mysofa.h"
 
+#if HAVE_CXXMODULES
+import gsl;
+#else
+#include "gsl/gsl"
+#endif
+
 
 namespace {
 
@@ -59,7 +68,7 @@ using uint = unsigned int;
  * possible.  Those sets that contain purely random measurements or use
  * different major axes will fail.
  */
-auto PrepareLayout(const al::span<const float> xyzs, HrirDataT *hData) -> bool
+auto PrepareLayout(const std::span<const float> xyzs, HrirDataT *hData) -> bool
 {
     fmt::println("Detecting compatible layout...");
 
@@ -75,7 +84,8 @@ auto PrepareLayout(const al::span<const float> xyzs, HrirDataT *hData) -> bool
     auto azCounts = std::vector<std::array<uint,MAX_EV_COUNT>>(MAX_FD_COUNT);
     for(auto &azs : azCounts) azs.fill(0u);
 
-    uint fi{0u}, ir_total{0u};
+    auto fi = 0u;
+    auto ir_total = 0u;
     for(const auto &field : fds)
     {
         distances[fi] = field.mDistance;
@@ -92,23 +102,23 @@ auto PrepareLayout(const al::span<const float> xyzs, HrirDataT *hData) -> bool
         ++fi;
     }
     fmt::println("Using {} of {} IRs.", ir_total, xyzs.size()/3);
-    const auto azs = al::span{azCounts}.first<MAX_FD_COUNT>();
-    return PrepareHrirData(al::span{distances}.first(fi), evCounts, azs, hData);
+    const auto azs = std::span{azCounts}.first<MAX_FD_COUNT>();
+    return PrepareHrirData(std::span{distances}.first(fi), evCounts, azs, hData);
 }
 
-float GetSampleRate(MYSOFA_HRTF *sofaHrtf)
+auto GetSampleRate(MYSOFA_HRTF const *const sofaHrtf) -> f32
 {
-    const char *srate_dim{nullptr};
-    const char *srate_units{nullptr};
-    MYSOFA_ARRAY *srate_array{&sofaHrtf->DataSamplingRate};
-    MYSOFA_ATTRIBUTE *srate_attrs{srate_array->attributes};
+    auto *srate_dim = gsl::czstring{};
+    auto *srate_units = gsl::czstring{};
+    auto const &srate_array = sofaHrtf->DataSamplingRate;
+    auto const *srate_attrs = srate_array.attributes;
     while(srate_attrs)
     {
         if("DIMENSION_LIST"sv == srate_attrs->name)
         {
             if(srate_dim)
             {
-                fmt::println(stderr, "Duplicate SampleRate.DIMENSION_LIST");
+                fmt::println(std::cerr, "Duplicate SampleRate.DIMENSION_LIST");
                 return 0.0f;
             }
             srate_dim = srate_attrs->value;
@@ -117,76 +127,75 @@ float GetSampleRate(MYSOFA_HRTF *sofaHrtf)
         {
             if(srate_units)
             {
-                fmt::println(stderr, "Duplicate SampleRate.Units");
+                fmt::println(std::cerr, "Duplicate SampleRate.Units");
                 return 0.0f;
             }
             srate_units = srate_attrs->value;
         }
         else
-            fmt::println(stderr, "Unexpected sample rate attribute: {} = {}", srate_attrs->name,
-                srate_attrs->value);
+            fmt::println(std::cerr, "Unexpected sample rate attribute: {} = {}", srate_attrs->name,
+                srate_attrs->value ? srate_attrs->value : "<null>");
         srate_attrs = srate_attrs->next;
     }
     if(!srate_dim)
     {
-        fmt::println(stderr, "Missing sample rate dimensions");
+        fmt::println(std::cerr, "Missing sample rate dimensions");
         return 0.0f;
     }
     if(srate_dim != "I"sv)
     {
-        fmt::println(stderr, "Unsupported sample rate dimensions: {}", srate_dim);
+        fmt::println(std::cerr, "Unsupported sample rate dimensions: {}", srate_dim);
         return 0.0f;
     }
     if(!srate_units)
     {
-        fmt::println(stderr, "Missing sample rate unit type");
+        fmt::println(std::cerr, "Missing sample rate unit type");
         return 0.0f;
     }
     if(srate_units != "hertz"sv)
     {
-        fmt::println(stderr, "Unsupported sample rate unit type: {}", srate_units);
+        fmt::println(std::cerr, "Unsupported sample rate unit type: {}", srate_units);
         return 0.0f;
     }
     /* I dimensions guarantees 1 element, so just extract it. */
-    const auto values = al::span{srate_array->values, sofaHrtf->I};
+    auto const values = std::span{srate_array.values, sofaHrtf->I};
     if(values[0] < float{MIN_RATE} || values[0] > float{MAX_RATE})
     {
-        fmt::println(stderr, "Sample rate out of range: {:f} (expected {} to {})", values[0],
+        fmt::println(std::cerr, "Sample rate out of range: {:f} (expected {} to {})", values[0],
             MIN_RATE, MAX_RATE);
         return 0.0f;
     }
     return values[0];
 }
 
-enum class DelayType : uint8_t {
+enum class DelayType : u8 {
     None,
     I_R, /* [1][Channels] */
     M_R, /* [HRIRs][Channels] */
 };
-auto PrepareDelay(MYSOFA_HRTF *sofaHrtf) -> std::optional<DelayType>
+auto PrepareDelay(MYSOFA_HRTF const *const sofaHrtf) -> std::optional<DelayType>
 {
-    const char *delay_dim{nullptr};
-    MYSOFA_ARRAY *delay_array{&sofaHrtf->DataDelay};
-    MYSOFA_ATTRIBUTE *delay_attrs{delay_array->attributes};
+    auto *delay_dim = gsl::czstring{};
+    auto const *delay_attrs = sofaHrtf->DataDelay.attributes;
     while(delay_attrs)
     {
         if("DIMENSION_LIST"sv == delay_attrs->name)
         {
             if(delay_dim)
             {
-                fmt::println(stderr, "Duplicate Delay.DIMENSION_LIST");
+                fmt::println(std::cerr, "Duplicate Delay.DIMENSION_LIST");
                 return std::nullopt;
             }
             delay_dim = delay_attrs->value;
         }
         else
-            fmt::println(stderr, "Unexpected delay attribute: {} = {}", delay_attrs->name,
+            fmt::println(std::cerr, "Unexpected delay attribute: {} = {}", delay_attrs->name,
                 delay_attrs->value ? delay_attrs->value : "<null>");
         delay_attrs = delay_attrs->next;
     }
     if(!delay_dim)
     {
-        fmt::println(stderr, "Missing delay dimensions");
+        fmt::println(std::cerr, "Missing delay dimensions");
         return DelayType::None;
     }
     if(delay_dim == "I,R"sv)
@@ -194,39 +203,38 @@ auto PrepareDelay(MYSOFA_HRTF *sofaHrtf) -> std::optional<DelayType>
     if(delay_dim == "M,R"sv)
         return DelayType::M_R;
 
-    fmt::println(stderr, "Unsupported delay dimensions: {}", delay_dim);
+    fmt::println(std::cerr, "Unsupported delay dimensions: {}", delay_dim);
     return std::nullopt;
 }
 
-bool CheckIrData(MYSOFA_HRTF *sofaHrtf)
+auto CheckIrData(MYSOFA_HRTF const *const sofaHrtf) -> bool
 {
-    const char *ir_dim{nullptr};
-    MYSOFA_ARRAY *ir_array{&sofaHrtf->DataIR};
-    MYSOFA_ATTRIBUTE *ir_attrs{ir_array->attributes};
+    auto *ir_dim = gsl::czstring{};
+    auto const *ir_attrs = sofaHrtf->DataIR.attributes;
     while(ir_attrs)
     {
         if("DIMENSION_LIST"sv == ir_attrs->name)
         {
             if(ir_dim)
             {
-                fmt::println(stderr, "Duplicate IR.DIMENSION_LIST");
+                fmt::println(std::cerr, "Duplicate IR.DIMENSION_LIST");
                 return false;
             }
             ir_dim = ir_attrs->value;
         }
         else
-            fmt::println(stderr, "Unexpected IR attribute: {} = {}", ir_attrs->name,
+            fmt::println(std::cerr, "Unexpected IR attribute: {} = {}", ir_attrs->name,
                 ir_attrs->value ? ir_attrs->value : "<null>");
         ir_attrs = ir_attrs->next;
     }
     if(!ir_dim)
     {
-        fmt::println(stderr, "Missing IR dimensions");
+        fmt::println(std::cerr, "Missing IR dimensions");
         return false;
     }
     if(ir_dim != "M,R,N"sv)
     {
-        fmt::println(stderr, "Unsupported IR dimensions: {}", ir_dim);
+        fmt::println(std::cerr, "Unsupported IR dimensions: {}", ir_dim);
         return false;
     }
     return true;
@@ -235,22 +243,21 @@ bool CheckIrData(MYSOFA_HRTF *sofaHrtf)
 
 /* Calculate the onset time of a HRIR. */
 constexpr int OnsetRateMultiple{10};
-auto CalcHrirOnset(PPhaseResampler &rs, const uint rate, al::span<double> upsampled,
-    const al::span<const double> hrir) -> double
+auto CalcHrirOnset(PPhaseResampler &rs, const uint rate, std::span<double> upsampled,
+    const std::span<const double> hrir) -> double
 {
     rs.process(hrir, upsampled);
 
-    auto abs_lt = [](const double lhs, const double rhs) -> bool
-    { return std::abs(lhs) < std::abs(rhs); };
-    auto iter = std::max_element(upsampled.cbegin(), upsampled.cend(), abs_lt);
-    return static_cast<double>(std::distance(upsampled.cbegin(), iter)) /
+    static constexpr auto make_abs = [](const double value) { return std::abs(value); };
+    const auto iter = std::ranges::max_element(upsampled, std::less{}, make_abs);
+    return static_cast<double>(std::distance(upsampled.begin(), iter)) /
         (double{OnsetRateMultiple}*rate);
 }
 
 /* Calculate the magnitude response of a HRIR. */
-void CalcHrirMagnitude(const uint points, al::span<complex_d> h, const al::span<double> hrir)
+void CalcHrirMagnitude(const uint points, std::span<complex_d> h, const std::span<double> hrir)
 {
-    auto iter = std::copy_n(hrir.cbegin(), points, h.begin());
+    auto iter = std::copy_n(hrir.begin(), points, h.begin());
     std::fill(iter, h.end(), complex_d{0.0, 0.0});
 
     forward_fft(h);
@@ -260,13 +267,13 @@ void CalcHrirMagnitude(const uint points, al::span<complex_d> h, const al::span<
 bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType delayType,
     const uint outRate)
 {
-    std::atomic<uint> loaded_count{0u};
+    auto loaded_count = std::atomic{0u};
 
     auto load_proc = [sofaHrtf,hData,delayType,outRate,&loaded_count]() -> bool
     {
         const uint channels{(hData->mChannelType == CT_STEREO) ? 2u : 1u};
         hData->mHrirsBase.resize(channels * size_t{hData->mIrCount} * hData->mIrSize, 0.0);
-        const auto hrirs = al::span{hData->mHrirsBase};
+        const auto hrirs = std::span{hData->mHrirsBase};
 
         std::vector<double> restmp;
         std::optional<PPhaseResampler> resampler;
@@ -276,8 +283,8 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
             restmp.resize(sofaHrtf->N);
         }
 
-        const auto srcPosValues = al::span{sofaHrtf->SourcePosition.values, sofaHrtf->M*3_uz};
-        const auto irValues = al::span{sofaHrtf->DataIR.values,
+        const auto srcPosValues = std::span{sofaHrtf->SourcePosition.values, sofaHrtf->M*3_uz};
+        const auto irValues = std::span{sofaHrtf->DataIR.values,
             size_t{sofaHrtf->M}*sofaHrtf->R*sofaHrtf->N};
         for(uint si{0u};si < sofaHrtf->M;++si)
         {
@@ -292,10 +299,9 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
             else
                 aer[0] = std::fmod(360.0f - aer[0], 360.0f);
 
-            auto field = std::find_if(hData->mFds.cbegin(), hData->mFds.cend(),
-                [&aer](const HrirFdT &fld) -> bool
-                { return (std::abs(aer[2] - fld.mDistance) < 0.001); });
-            if(field == hData->mFds.cend())
+            auto field = std::ranges::find_if(hData->mFds, [&aer](const HrirFdT &fld) -> bool
+            { return (std::abs(aer[2] - fld.mDistance) < 0.001); });
+            if(field == hData->mFds.end())
                 continue;
 
             const double evscale{180.0 / static_cast<double>(field->mEvs.size()-1)};
@@ -314,7 +320,7 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
             HrirAzT &azd = field->mEvs[ei].mAzs[ai];
             if(!azd.mIrs[0].empty())
             {
-                fmt::println(stderr, "\nMultiple measurements near [ a={:f}, e={:f}, r={:f} ].",
+                fmt::println(std::cerr, "\nMultiple measurements near [ a={:f}, e={:f}, r={:f} ].",
                     aer[0], aer[1], aer[2]);
                 return false;
             }
@@ -326,10 +332,10 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
                 const auto ir = irValues.subspan((size_t{si}*sofaHrtf->R + ti)*sofaHrtf->N,
                     sofaHrtf->N);
                 if(!resampler)
-                    std::copy_n(ir.cbegin(), ir.size(), azd.mIrs[ti].begin());
+                    std::copy_n(ir.begin(), ir.size(), azd.mIrs[ti].begin());
                 else
                 {
-                    std::copy_n(ir.cbegin(), ir.size(), restmp.begin());
+                    std::copy_n(ir.begin(), ir.size(), restmp.begin());
                     resampler->process(restmp, azd.mIrs[ti]);
                 }
             }
@@ -337,14 +343,14 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
             /* Include any per-channel or per-HRIR delays. */
             if(delayType == DelayType::I_R)
             {
-                const auto delayValues = al::span{sofaHrtf->DataDelay.values,
+                const auto delayValues = std::span{sofaHrtf->DataDelay.values,
                     size_t{sofaHrtf->I}*sofaHrtf->R};
                 for(uint ti{0u};ti < channels;++ti)
                     azd.mDelays[ti] = delayValues[ti] / static_cast<float>(hData->mIrRate);
             }
             else if(delayType == DelayType::M_R)
             {
-                const auto delayValues = al::span{sofaHrtf->DataDelay.values,
+                const auto delayValues = std::span{sofaHrtf->DataDelay.values,
                     size_t{sofaHrtf->M}*sofaHrtf->R};
                 for(uint ti{0u};ti < channels;++ti)
                     azd.mDelays[ti] = delayValues[si*sofaHrtf->R + ti] /
@@ -367,7 +373,7 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
     do {
         load_status = load_future.wait_for(std::chrono::milliseconds{50});
         fmt::print("\rLoading HRIRs... {} of {}", loaded_count.load(), sofaHrtf->M);
-        fflush(stdout);
+        std::cout.flush();
     } while(load_status != std::future_status::ready);
     fmt::println("");
     return load_future.get();
@@ -381,9 +387,9 @@ bool LoadResponses(MYSOFA_HRTF *sofaHrtf, HrirDataT *hData, const DelayType dela
 struct MagCalculator {
     const uint mFftSize{};
     const uint mIrPoints{};
-    std::vector<al::span<double>> mIrs;
-    std::atomic<size_t> mCurrent{};
-    std::atomic<size_t> mDone{};
+    std::vector<std::span<double>> mIrs;
+    std::atomic<size_t> mCurrent;
+    std::atomic<size_t> mDone;
 
     MagCalculator(const uint fftsize, const uint irpoints) : mFftSize{fftsize}, mIrPoints{irpoints}
     { }
@@ -431,20 +437,20 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
     /* NOTE: Some valid SOFA files are failing this check. */
     err = mysofa_check(sofaHrtf.get());
     if(err != MYSOFA_OK)
-        fmt::println(stderr, "Warning: Supposedly malformed source file '{}': {} ({})", filename,
-            SofaErrorStr(err), err);
+        fmt::println(std::cerr, "Warning: Supposedly malformed source file '{}': {} ({})",
+            filename, SofaErrorStr(err), err);
 
     mysofa_tocartesian(sofaHrtf.get());
 
     /* Make sure emitter and receiver counts are sane. */
     if(sofaHrtf->E != 1)
     {
-        fmt::println(stderr, "{} emitters not supported", sofaHrtf->E);
+        fmt::println(std::cerr, "{} emitters not supported", sofaHrtf->E);
         return false;
     }
     if(sofaHrtf->R > 2 || sofaHrtf->R < 1)
     {
-        fmt::println(stderr, "{} receivers not supported", sofaHrtf->R);
+        fmt::println(std::cerr, "{} receivers not supported", sofaHrtf->R);
         return false;
     }
     /* Assume R=2 is a stereo measurement, and R=1 is mono left-ear-only. */
@@ -456,14 +462,14 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
     /* Check and set the FFT and IR size. */
     if(sofaHrtf->N > fftSize)
     {
-        fmt::println(stderr, "Sample points exceeds the FFT size ({} > {}).", sofaHrtf->N,
+        fmt::println(std::cerr, "Sample points exceeds the FFT size ({} > {}).", sofaHrtf->N,
             fftSize);
         return false;
     }
     if(sofaHrtf->N < truncSize)
     {
-        fmt::println(stderr, "Sample points is below the truncation size ({} < {}).", sofaHrtf->N,
-            truncSize);
+        fmt::println(std::cerr, "Sample points is below the truncation size ({} < {}).",
+            sofaHrtf->N, truncSize);
         return false;
     }
     hData->mIrPoints = sofaHrtf->N;
@@ -483,7 +489,7 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
 
     if(!CheckIrData(sofaHrtf.get()))
         return false;
-    if(!PrepareLayout(al::span{sofaHrtf->SourcePosition.values, sofaHrtf->M*3_uz}, hData))
+    if(!PrepareLayout(std::span{sofaHrtf->SourcePosition.values, sofaHrtf->M*3_uz}, hData))
         return false;
     if(!LoadResponses(sofaHrtf.get(), hData, *delayType, outRate))
         return false;
@@ -505,7 +511,7 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
         }
         if(ei >= hData->mFds[fi].mEvs.size())
         {
-            fmt::println(stderr, "Missing source references [ {}, *, * ].", fi);
+            fmt::println(std::cerr, "Missing source references [{}, *, *].", fi);
             return false;
         }
         hData->mFds[fi].mEvStart = ei;
@@ -516,7 +522,7 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
                 HrirAzT &azd = hData->mFds[fi].mEvs[ei].mAzs[ai];
                 if(azd.mIrs[0].empty())
                 {
-                    fmt::println(stderr, "Missing source reference [ {}, {}, {} ].", fi, ei, ai);
+                    fmt::println(std::cerr, "Missing source reference [{}, {}, {}].", fi, ei, ai);
                     return false;
                 }
             }
@@ -526,7 +532,7 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
 
     size_t hrir_total{0};
     const uint channels{(hData->mChannelType == CT_STEREO) ? 2u : 1u};
-    const auto hrirs = al::span{hData->mHrirsBase};
+    const auto hrirs = std::span{hData->mHrirsBase};
     for(uint fi{0u};fi < hData->mFds.size();fi++)
     {
         for(uint ei{0u};ei < hData->mFds[fi].mEvStart;ei++)
@@ -576,7 +582,7 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
     do {
         load_status = load_future.wait_for(std::chrono::milliseconds{50});
         fmt::print("\rCalculating HRIR onsets... {} of {}", hrir_done.load(), hrir_total);
-        fflush(stdout);
+        std::cout.flush();
     } while(load_status != std::future_status::ready);
     fmt::println("");
     if(!load_future.get())
@@ -598,14 +604,14 @@ bool LoadSofaFile(const std::string_view filename, const uint numThreads, const
     std::vector<std::thread> thrds;
     thrds.reserve(numThreads);
     for(size_t i{0};i < numThreads;++i)
-        thrds.emplace_back(std::mem_fn(&MagCalculator::Worker), &calculator);
+        thrds.emplace_back(&MagCalculator::Worker, &calculator);
     size_t count;
     do {
         std::this_thread::sleep_for(std::chrono::milliseconds{50});
         count = calculator.mDone.load();
 
         fmt::print("\rCalculating HRIR magnitudes... {} of {}", count, calculator.mIrs.size());
-        fflush(stdout);
+        std::cout.flush();
     } while(count != calculator.mIrs.size());
     fmt::println("");
 
